# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006-2008 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.
#
# Author: Alessandro Decina <alessandro@fluendo.com>

import dbus
import dbus.service
import dbus.glib
import gobject
from twisted.internet import defer, task

from elisa.core import common
from elisa.core.components.service_provider import ServiceProvider
from elisa.core.application import ComponentsLoadedMessage
from elisa.plugins.database.models import Artist, MusicAlbum, MusicTrack, \
        PhotoAlbum, Image, File
from storm.expr import Desc, Asc
 
def result_set_order_offset_and_limit(result_set, field, offset, limit):
    result_set.order_by(field).config(offset=offset, limit=limit)
    return result_set

def result_set_first(result_set):
    return result_set.first()

def result_set_one(result_set):
    return result_set.one()

def result_set_any(result_set):
    return result_set.any()

def result_set_all(result_set):
    return result_set.all()

def dbus_errback(failure, errback):
    errback(failure.value)
    
def build_last_track_played_tuple(music_track):
    def get_artist(album):
        def build_tuple(artist):
            return (artist.name, album.name,
                    music_track.title, music_track.file_path)

        dfr = music_track.artists.find()
        dfr.addCallback(result_set_any)
        dfr.addCallback(build_tuple)

        return dfr

    if music_track is None:
        raise MusicTrackNotFoundError()

    dfr = music_track.album
    dfr.addCallback(get_artist)

    return dfr

def build_last_picture_played_tuple(picture):
    def build_tuple(album):
        return album.name, picture.file_path

    if picture is None:
        raise PictureNotFoundError()

    dfr = picture.album
    dfr.addCallback(build_tuple)

    return dfr

class DatabaseError(Exception):
    pass

class ArtistNotFoundError(DatabaseError):
    pass

class MusicAlbumNotFoundError(DatabaseError):
    pass

class MusicTrackNotFoundError(DatabaseError):
    pass

class PhotoAlbumNotFoundError(DatabaseError):
    pass

class PictureNotFoundError(DatabaseError):
    pass

class MusicImpl(object):
    def _build_artists_albums_list(self, albums):
        def find_album_tracks_iterator(albums, artists_albums_set):
            for album in albums:
                dfr = album.tracks.find()
                dfr.addCallback(result_set_all)
                dfr.addCallback(self._find_tracks_artists,
                        album, artists_albums_set)
                yield dfr

        def find_album_tracks_iterator_cb(iterator, artists_albums_set):
            return list(artists_albums_set)

        artists_albums_set = set()
        dfr = task.coiterate(find_album_tracks_iterator(albums,
                artists_albums_set))
        dfr.addCallback(find_album_tracks_iterator_cb, artists_albums_set)
        return dfr

    def _find_tracks_artists(self, tracks, album, artists_albums_set):
        def fill_artists_albums_set(artists, album, artists_albums_set):
            for artist in artists:
                artists_albums_set.add((artist.name, album.name))

        def find_tracks_artists_iterator(tracks, album, artists_albums_set):
            for track in tracks:
                dfr = track.artists.find()
                dfr.addCallback(result_set_all)
                dfr.addCallback(fill_artists_albums_set, album, artists_albums_set)

                yield dfr

        def find_tracks_artists_iterator_cb(iterator, artists_albums_set):
            return list(artists_albums_set)

        dfr = task.coiterate(find_tracks_artists_iterator(tracks, album,
                artists_albums_set))
        dfr.addCallback(find_tracks_artists_iterator_cb, artists_albums_set)

        return dfr

    def get_albums(self, offset, count):
        if offset < 0 or count <= 0:
            return defer.fail(IndexError())

        dfr = common.application.store.find(MusicAlbum)
        dfr.addCallback(result_set_order_offset_and_limit, MusicAlbum.name, offset, count)
        dfr.addCallback(result_set_all)
        dfr.addCallback(self._build_artists_albums_list)

        return dfr

    def get_album_artwork(self, artist_name, album_name):
        def get_album(music_track):
            if music_track is None:
                raise MusicAlbumNotFoundError()

            return music_track.album

        def get_album_cover_uri(album):
            return album.cover_uri

        def find_cover_art(artist):
            if artist is None:
                raise ArtistNotFoundError()

            dfr = artist.tracks.find(MusicTrack.album_name == album_name)
            dfr.addCallback(result_set_any)
            dfr.addCallback(get_album)
            dfr.addCallback(get_album_cover_uri)

            return dfr

        dfr = common.application.store.find(Artist, Artist.name == artist_name)
        dfr.addCallback(result_set_one)
        dfr.addCallback(find_cover_art)
    
        return dfr

    def get_album_tracks(self, artist_name, album_name):
        def build_track_list(tracks):
            if not tracks:
                raise MusicAlbumNotFoundError()

            return [(track.title, track.file_path) for track in tracks]

        def find_tracks(artist):
            if artist is None:
                raise ArtistNotFoundError()

            dfr = artist.tracks.find(MusicTrack.album_name == album_name)
            dfr.addCallback(result_set_all)
            dfr.addCallback(build_track_list)

            return dfr

        dfr = common.application.store.find(Artist, Artist.name == artist_name)
        dfr.addCallback(result_set_one)
        dfr.addCallback(find_tracks)

        return dfr

    def get_last_played_track(self):
        dfr = common.application.store.find(MusicTrack,
                MusicTrack.file_path == File.path,
                File.last_played != 0)
        dfr.addCallback(result_set_order_offset_and_limit,
                Desc(File.last_played), offset=None, limit=None)
        dfr.addCallback(result_set_first)
        dfr.addCallback(build_last_track_played_tuple)

        return dfr

class Music(dbus.service.Object):
    interface = 'com.fluendo.Elisa.Plugins.Database.Music'

    def __init__(self, *args, **kw):
        dbus.service.Object.__init__(self, *args, **kw)
        self.impl = MusicImpl()

    @dbus.service.method(dbus_interface=interface,
            in_signature='ii', out_signature='aas',
            async_callbacks=('callback', 'errback')) 
    def get_albums(self, offset, count, callback, errback):
        dfr = self.impl.get_albums(offset, count)
        dfr.addCallback(callback)
        dfr.addErrback(dbus_errback, errback)

    @dbus.service.method(dbus_interface=interface,
            in_signature='ss', out_signature='s',
            async_callbacks=('callback', 'errback')) 
    def get_album_artwork(self, artist_name, album_name, callback, errback):
        dfr = self.impl.get_album_artwork(artist_name, album_name)
        dfr.addCallback(callback)
        dfr.addErrback(dbus_errback, errback)
    
    @dbus.service.method(dbus_interface=interface,
            in_signature='ss', out_signature='aas',
            async_callbacks=('callback', 'errback')) 
    def get_album_tracks(self, artist_name, album_name, callback, errback):
        dfr = self.impl.get_album_tracks(artist_name, album_name)
        dfr.addCallback(callback)
        dfr.addErrback(dbus_errback, errback)

    @dbus.service.method(dbus_interface=interface,
            in_signature='', out_signature='as',
            async_callbacks=('callback', 'errback')) 
    def get_last_played_track(self, callback, errback):
        dfr = self.impl.get_last_played_track()
        dfr.addCallback(callback)
        dfr.addErrback(dbus_errback, errback)

    @dbus.service.signal(dbus_interface=interface, signature='ssss')
    def last_played_track(self, artist_name, album_name,
            track_title, track_path):
        # don't do anything, just so that the signal is emitted
        pass

class PhotoImpl(object):
    def get_albums(self, offset, count):
        if offset < 0 or count <= 0:
            return defer.fail(IndexError())
        
        def build_album_list(albums):
            return [album.name for album in albums]

        dfr = common.application.store.find(PhotoAlbum)
        dfr.addCallback(result_set_order_offset_and_limit, PhotoAlbum.name, offset, count)
        dfr.addCallback(result_set_all)
        dfr.addCallback(build_album_list)

        return dfr

    def get_album_pictures(self, album_name):
        def build_pictures_list(pictures):
            return [picture.file_path for picture in pictures]

        def get_album_pictures(album):
            if album is None:
                raise PhotoAlbumNotFoundError()

            dfr = album.photos.find()
            dfr.addCallback(result_set_all)
            dfr.addCallback(build_pictures_list)

            return dfr

        dfr = common.application.store.find(PhotoAlbum,
                PhotoAlbum.name == album_name)
        dfr.addCallback(result_set_one)
        dfr.addCallback(get_album_pictures)

        return dfr
    
    def get_last_displayed_picture(self):
        dfr = common.application.store.find(Image,
                Image.file_path == File.path,
                File.last_played != 0)
        dfr.addCallback(result_set_order_offset_and_limit,
                Desc(File.last_played), offset=None, limit=None)
        dfr.addCallback(result_set_first)
        dfr.addCallback(build_last_picture_played_tuple)

        return dfr

class Photo(dbus.service.Object):
    interface = 'com.fluendo.Elisa.Plugins.Database.Photo'
    
    def __init__(self, *args, **kw):
        dbus.service.Object.__init__(self, *args, **kw)
        self.impl = PhotoImpl()

    @dbus.service.method(dbus_interface=interface,
            in_signature='ii', out_signature='as',
            async_callbacks=('callback', 'errback')) 
    def get_albums(self, offset, count, callback, errback):
        dfr = self.impl.get_albums(offset, count)
        dfr.addCallback(callback)
        dfr.addErrback(dbus_errback, errback)

    @dbus.service.method(dbus_interface=interface,
            in_signature='s', out_signature='as',
            async_callbacks=('callback', 'errback')) 
    def get_album_pictures(self, album_name, callback, errback):
        dfr = self.impl.get_album_pictures(album_name)
        dfr.addCallback(callback)
        dfr.addErrback(dbus_errback, errback)
    
    @dbus.service.method(dbus_interface=interface,
            in_signature='', out_signature='as',
            async_callbacks=('callback', 'errback')) 
    def get_last_displayed_picture(self, callback, errback):
        dfr = self.impl.get_last_displayed_picture()
        dfr.addCallback(callback)
        dfr.addErrback(dbus_errback, errback)

    @dbus.service.signal(dbus_interface=interface, signature='ss')
    def last_displayed_picture(self, photo_album_name, picture_path):
        # don't do anything, just so that the signal is emitted
        pass

class DatabaseDBusServiceProvider(ServiceProvider):
    def start(self): 
        self._last_played_track_path = None
        self._last_displayed_picture_path = None
        
        bus = dbus.SessionBus()
        
        # get known names
        self.bus_name = \
                dbus.service.BusName('com.fluendo.Elisa.Plugins.Database', bus)
    
        # create the objects we export
        self.photo = Photo(bus, '/com/fluendo/Elisa/Plugins/Database/Photo',
                self.bus_name)
        self.music = Music(bus, '/com/fluendo/Elisa/Plugins/Database/Music',
                self.bus_name)

        self._register_player_status_changes()

        return defer.succeed(self)

    def stop(self):
        bus = dbus.SessionBus()

        self.photo.remove_from_connection(bus,
                '/com/fluendo/Elisa/Plugins/Database/Photo')
        self.music.remove_from_connection(bus,
                '/com/fluendo/Elisa/Plugins/Database/Music')

        # BusName implements __del__, eew
        del self.bus_name

        return defer.succeed(self)

    def _register_player_status_changes(self):
        bus = common.application.bus
        bus.register(self._components_loaded, ComponentsLoadedMessage)

    def _components_loaded(self, msg, sender):
        # FIXME: we need an API to retrieve the frontend
        frontends = common.application.interface_controller.frontends
        frontend = frontends.values()[0]
        controllers = frontend.retrieve_controllers('/poblesec') 
        poblesec = controllers[0]
        music_player = \
                frontend.retrieve_controllers('/poblesec/music_player')[0]
        music_player.player.connect('status-changed',self._music_player_status_cb)
        slideshow_player = \
                frontend.retrieve_controllers('/poblesec/slideshow_player')[0]
        slideshow_player.player.connect('current-index-changed',
                self._slideshow_player_current_index_changed_cb)

    def _music_player_status_cb(self, player, status):
        if status != player.PLAYING:
            return defer.succeed(None)

        current_track_path = player.playlist[player.current_index].uri.path
        if self._last_played_track_path is None or \
                current_track_path != self._last_played_track_path:
            
            self._last_played_track_path = current_track_path

            dfr = common.application.store.find(MusicTrack,
                    MusicTrack.file_path == current_track_path)
            dfr.addCallback(result_set_one)
            dfr.addCallback(build_last_track_played_tuple)
            dfr.addCallback(self._emit_last_played_track)

            return dfr

        return defer.succeed(None)
    
    def _emit_last_played_track(self, last_played_track_tuple):
        self.music.last_played_track(*last_played_track_tuple)

    def _slideshow_player_current_index_changed_cb(self,
            player, index, previous_index):
        current_picture_path = player.playlist[index].references[-1].path

        if self._last_displayed_picture_path is None or \
                current_picture_path != self._last_displayed_picture_path:
            self._last_displayed_picture_path = current_picture_path
           
            dfr = common.application.store.find(Image,
                    Image.file_path == current_picture_path)
            dfr.addCallback(result_set_one)
            dfr.addCallback(build_last_picture_played_tuple)
            dfr.addCallback(self._emit_last_displayed_picture)

            return dfr
        
        return defer.succeed(None)
    
    def _emit_last_displayed_picture(self, last_displayed_picture_tuple):
        self.photo.last_displayed_picture(*last_displayed_picture_tuple)

if __name__ == '__main__':
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
    loop = gobject.MainLoop()

    def service_provider_created_cb(service_provider):
        service_provider.start()

    dfr = DatabaseDBusServiceProvider.create({})
    dfr.addCallback(service_provider_created_cb)

    loop.run()
